import warnings
from functools import wraps
from typing import Any, Callable

from twisted.internet import defer, threads
from twisted.internet.defer import Deferred

from scrapy.exceptions import ScrapyDeprecationWarning


def deprecated(use_instead: Any = None) -> Callable:
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used."""

    def deco(func: Callable) -> Callable:
        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            message = f"Call to deprecated function {func.__name__}."
            if use_instead:
                message += f" Use {use_instead} instead."
            warnings.warn(message, category=ScrapyDeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapped

    if callable(use_instead):
        deco = deco(use_instead)
        use_instead = None
    return deco


def defers(func: Callable) -> Callable[..., Deferred]:
    """Decorator to make sure a function always returns a deferred"""

    @wraps(func)
    def wrapped(*a: Any, **kw: Any) -> Deferred:
        return defer.maybeDeferred(func, *a, **kw)

    return wrapped


def inthread(func: Callable) -> Callable[..., Deferred]:
    """Decorator to call a function in a thread and return a deferred with the
    result
    """

    @wraps(func)
    def wrapped(*a: Any, **kw: Any) -> Deferred:
        return threads.deferToThread(func, *a, **kw)

    return wrapped



import shutil
from pathlib import Path

from testfixtures import LogCapture
from twisted.internet import defer
from twisted.trial.unittest import TestCase
from w3lib.url import add_or_replace_parameter

from scrapy import signals
from scrapy.crawler import CrawlerRunner
from tests.mockserver import MockServer
from tests.spiders import SimpleSpider


class MediaDownloadSpider(SimpleSpider):
    name = "mediadownload"

    def _process_url(self, url):
        return url

    def parse(self, response):
        self.logger.info(response.headers)
        self.logger.info(response.text)
        item = {
            self.media_key: [],
            self.media_urls_key: [
                self._process_url(response.urljoin(href))
                for href in response.xpath(
                    '//table[thead/tr/th="Filename"]/tbody//a/@href'
                ).getall()
            ],
        }
        yield item


class BrokenLinksMediaDownloadSpider(MediaDownloadSpider):
    name = "brokenmedia"

    def _process_url(self, url):
        return url + ".foo"


class RedirectedMediaDownloadSpider(MediaDownloadSpider):
    name = "redirectedmedia"

    def _process_url(self, url):
        return add_or_replace_parameter(
            self.mockserver.url("/redirect-to"), "goto", url
        )


class FileDownloadCrawlTestCase(TestCase):
    pipeline_class = "scrapy.pipelines.files.FilesPipeline"
    store_setting_key = "FILES_STORE"
    media_key = "files"
    media_urls_key = "file_urls"
    expected_checksums = {
        "5547178b89448faf0015a13f904c936e",
        "c2281c83670e31d8aaab7cb642b824db",
        "ed3f6538dc15d4d9179dae57319edc5f",
    }

    def setUp(self):
        self.mockserver = MockServer()
        self.mockserver.__enter__()

        # prepare a directory for storing files
        self.tmpmediastore = Path(self.mktemp())
        self.tmpmediastore.mkdir()
        self.settings = {
            "REQUEST_FINGERPRINTER_IMPLEMENTATION": "2.7",
            "ITEM_PIPELINES": {self.pipeline_class: 1},
            self.store_setting_key: str(self.tmpmediastore),
        }
        self.runner = CrawlerRunner(self.settings)
        self.items = []

    def tearDown(self):
        shutil.rmtree(self.tmpmediastore)
        self.items = []
        self.mockserver.__exit__(None, None, None)

    def _on_item_scraped(self, item):
        self.items.append(item)

    def _create_crawler(self, spider_class, runner=None, **kwargs):
        if runner is None:
            runner = self.runner
        crawler = runner.create_crawler(spider_class, **kwargs)
        crawler.signals.connect(self._on_item_scraped, signals.item_scraped)
        return crawler

    def _assert_files_downloaded(self, items, logs):
        self.assertEqual(len(items), 1)
        self.assertIn(self.media_key, items[0])

        # check that logs show the expected number of successful file downloads
        file_dl_success = "File (downloaded): Downloaded file from"
        self.assertEqual(logs.count(file_dl_success), 3)

        # check that the images/files status is `downloaded`
        for item in items:
            for i in item[self.media_key]:
                self.assertEqual(i["status"], "downloaded")

        # check that the images/files checksums are what we know they should be
        if self.expected_checksums is not None:
            checksums = set(
                i["checksum"] for item in items for i in item[self.media_key]
            )
            self.assertEqual(checksums, self.expected_checksums)

        # check that the image files where actually written to the media store
        for item in items:
            for i in item[self.media_key]:
                self.assertTrue((self.tmpmediastore / i["path"]).exists())

    def _assert_files_download_failure(self, crawler, items, code, logs):
        # check that the item does NOT have the "images/files" field populated
        self.assertEqual(len(items), 1)
        self.assertIn(self.media_key, items[0])
        self.assertFalse(items[0][self.media_key])

        # check that there was 1 successful fetch and 3 other responses with non-200 code
        self.assertEqual(
            crawler.stats.get_value("downloader/request_method_count/GET"), 4
        )
        self.assertEqual(crawler.stats.get_value("downloader/response_count"), 4)
        self.assertEqual(
            crawler.stats.get_value("downloader/response_status_count/200"), 1
        )
        self.assertEqual(
            crawler.stats.get_value(f"downloader/response_status_count/{code}"), 3
        )

        # check that logs do show the failure on the file downloads
        file_dl_failure = f"File (code: {code}): Error downloading file from"
        self.assertEqual(logs.count(file_dl_failure), 3)

        # check that no files were written to the media store
        self.assertEqual([x for x in self.tmpmediastore.iterdir()], [])

    @defer.inlineCallbacks
    def test_download_media(self):
        crawler = self._create_crawler(MediaDownloadSpider)
        with LogCapture() as log:
            yield crawler.crawl(
                self.mockserver.url("/files/images/"),
                media_key=self.media_key,
                media_urls_key=self.media_urls_key,
            )
        self._assert_files_downloaded(self.items, str(log))

    @defer.inlineCallbacks
    def test_download_media_wrong_urls(self):
        crawler = self._create_crawler(BrokenLinksMediaDownloadSpider)
        with LogCapture() as log:
            yield crawler.crawl(
                self.mockserver.url("/files/images/"),
                media_key=self.media_key,
                media_urls_key=self.media_urls_key,
            )
        self._assert_files_download_failure(crawler, self.items, 404, str(log))

    @defer.inlineCallbacks
    def test_download_media_redirected_default_failure(self):
        crawler = self._create_crawler(RedirectedMediaDownloadSpider)
        with LogCapture() as log:
            yield crawler.crawl(
                self.mockserver.url("/files/images/"),
                media_key=self.media_key,
                media_urls_key=self.media_urls_key,
                mockserver=self.mockserver,
            )
        self._assert_files_download_failure(crawler, self.items, 302, str(log))

    @defer.inlineCallbacks
    def test_download_media_redirected_allowed(self):
        settings = dict(self.settings)
        settings.update({"MEDIA_ALLOW_REDIRECTS": True})
        runner = CrawlerRunner(settings)
        crawler = self._create_crawler(RedirectedMediaDownloadSpider, runner=runner)
        with LogCapture() as log:
            yield crawler.crawl(
                self.mockserver.url("/files/images/"),
                media_key=self.media_key,
                media_urls_key=self.media_urls_key,
                mockserver=self.mockserver,
            )
        self._assert_files_downloaded(self.items, str(log))
        self.assertEqual(
            crawler.stats.get_value("downloader/response_status_count/302"), 3
        )


try:
    from PIL import Image  # noqa: imported just to check for the import error
except ImportError:
    skip_pillow = (
        "Missing Python Imaging Library, install https://pypi.python.org/pypi/Pillow"
    )
else:
    skip_pillow = None


class ImageDownloadCrawlTestCase(FileDownloadCrawlTestCase):
    skip = skip_pillow

    pipeline_class = "scrapy.pipelines.images.ImagesPipeline"
    store_setting_key = "IMAGES_STORE"
    media_key = "images"
    media_urls_key = "image_urls"

    # somehow checksums for images are different for Python 3.3
    expected_checksums = None



import dataclasses
import unittest

import attr
from itemadapter import ItemAdapter
from itemloaders.processors import Compose, Identity, MapCompose, TakeFirst

from scrapy.http import HtmlResponse, Response
from scrapy.item import Field, Item
from scrapy.loader import ItemLoader
from scrapy.selector import Selector


# test items
class NameItem(Item):
    name = Field()


class TestItem(NameItem):
    url = Field()
    summary = Field()


class TestNestedItem(Item):
    name = Field()
    name_div = Field()
    name_value = Field()

    url = Field()
    image = Field()


@attr.s
class AttrsNameItem:
    name = attr.ib(default="")


@dataclasses.dataclass
class TestDataClass:
    name: list = dataclasses.field(default_factory=list)


# test item loaders
class NameItemLoader(ItemLoader):
    default_item_class = TestItem


class NestedItemLoader(ItemLoader):
    default_item_class = TestNestedItem


class TestItemLoader(NameItemLoader):
    name_in = MapCompose(lambda v: v.title())


class DefaultedItemLoader(NameItemLoader):
    default_input_processor = MapCompose(lambda v: v[:-1])


# test processors
def processor_with_args(value, other=None, loader_context=None):
    if "key" in loader_context:
        return loader_context["key"]
    return value


class BasicItemLoaderTest(unittest.TestCase):
    def test_add_value_on_unknown_field(self):
        il = TestItemLoader()
        self.assertRaises(KeyError, il.add_value, "wrong_field", ["lala", "lolo"])

    def test_load_item_using_default_loader(self):
        i = TestItem()
        i["summary"] = "lala"
        il = ItemLoader(item=i)
        il.add_value("name", "marta")
        item = il.load_item()
        assert item is i
        self.assertEqual(item["summary"], ["lala"])
        self.assertEqual(item["name"], ["marta"])

    def test_load_item_using_custom_loader(self):
        il = TestItemLoader()
        il.add_value("name", "marta")
        item = il.load_item()
        self.assertEqual(item["name"], ["Marta"])


class InitializationTestMixin:
    item_class = None

    def test_keep_single_value(self):
        """Loaded item should contain values from the initial item"""
        input_item = self.item_class(name="foo")
        il = ItemLoader(item=input_item)
        loaded_item = il.load_item()
        self.assertIsInstance(loaded_item, self.item_class)
        self.assertEqual(ItemAdapter(loaded_item).asdict(), {"name": ["foo"]})

    def test_keep_list(self):
        """Loaded item should contain values from the initial item"""
        input_item = self.item_class(name=["foo", "bar"])
        il = ItemLoader(item=input_item)
        loaded_item = il.load_item()
        self.assertIsInstance(loaded_item, self.item_class)
        self.assertEqual(ItemAdapter(loaded_item).asdict(), {"name": ["foo", "bar"]})

    def test_add_value_singlevalue_singlevalue(self):
        """Values added after initialization should be appended"""
        input_item = self.item_class(name="foo")
        il = ItemLoader(item=input_item)
        il.add_value("name", "bar")
        loaded_item = il.load_item()
        self.assertIsInstance(loaded_item, self.item_class)
        self.assertEqual(ItemAdapter(loaded_item).asdict(), {"name": ["foo", "bar"]})

    def test_add_value_singlevalue_list(self):
        """Values added after initialization should be appended"""
        input_item = self.item_class(name="foo")
        il = ItemLoader(item=input_item)
        il.add_value("name", ["item", "loader"])
        loaded_item = il.load_item()
        self.assertIsInstance(loaded_item, self.item_class)
        self.assertEqual(
            ItemAdapter(loaded_item).asdict(), {"name": ["foo", "item", "loader"]}
        )

    def test_add_value_list_singlevalue(self):
        """Values added after initialization should be appended"""
        input_item = self.item_class(name=["foo", "bar"])
        il = ItemLoader(item=input_item)
        il.add_value("name", "qwerty")
        loaded_item = il.load_item()
        self.assertIsInstance(loaded_item, self.item_class)
        self.assertEqual(
            ItemAdapter(loaded_item).asdict(), {"name": ["foo", "bar", "qwerty"]}
        )

    def test_add_value_list_list(self):
        """Values added after initialization should be appended"""
        input_item = self.item_class(name=["foo", "bar"])
        il = ItemLoader(item=input_item)
        il.add_value("name", ["item", "loader"])
        loaded_item = il.load_item()
        self.assertIsInstance(loaded_item, self.item_class)
        self.assertEqual(
            ItemAdapter(loaded_item).asdict(),
            {"name": ["foo", "bar", "item", "loader"]},
        )

    def test_get_output_value_singlevalue(self):
        """Getting output value must not remove value from item"""
        input_item = self.item_class(name="foo")
        il = ItemLoader(item=input_item)
        self.assertEqual(il.get_output_value("name"), ["foo"])
        loaded_item = il.load_item()
        self.assertIsInstance(loaded_item, self.item_class)
        self.assertEqual(ItemAdapter(loaded_item).asdict(), dict({"name": ["foo"]}))

    def test_get_output_value_list(self):
        """Getting output value must not remove value from item"""
        input_item = self.item_class(name=["foo", "bar"])
        il = ItemLoader(item=input_item)
        self.assertEqual(il.get_output_value("name"), ["foo", "bar"])
        loaded_item = il.load_item()
        self.assertIsInstance(loaded_item, self.item_class)
        self.assertEqual(
            ItemAdapter(loaded_item).asdict(), dict({"name": ["foo", "bar"]})
        )

    def test_values_single(self):
        """Values from initial item must be added to loader._values"""
        input_item = self.item_class(name="foo")
        il = ItemLoader(item=input_item)
        self.assertEqual(il._values.get("name"), ["foo"])

    def test_values_list(self):
        """Values from initial item must be added to loader._values"""
        input_item = self.item_class(name=["foo", "bar"])
        il = ItemLoader(item=input_item)
        self.assertEqual(il._values.get("name"), ["foo", "bar"])


class InitializationFromDictTest(InitializationTestMixin, unittest.TestCase):
    item_class = dict


class InitializationFromItemTest(InitializationTestMixin, unittest.TestCase):
    item_class = NameItem


class InitializationFromAttrsItemTest(InitializationTestMixin, unittest.TestCase):
    item_class = AttrsNameItem


class InitializationFromDataClassTest(InitializationTestMixin, unittest.TestCase):
    item_class = TestDataClass


class BaseNoInputReprocessingLoader(ItemLoader):
    title_in = MapCompose(str.upper)
    title_out = TakeFirst()


class NoInputReprocessingItem(Item):
    title = Field()


class NoInputReprocessingItemLoader(BaseNoInputReprocessingLoader):
    default_item_class = NoInputReprocessingItem


class NoInputReprocessingFromItemTest(unittest.TestCase):
    """
    Loaders initialized from loaded items must not reprocess fields (Item instances)
    """

    def test_avoid_reprocessing_with_initial_values_single(self):
        il = NoInputReprocessingItemLoader(item=NoInputReprocessingItem(title="foo"))
        il_loaded = il.load_item()
        self.assertEqual(il_loaded, {"title": "foo"})
        self.assertEqual(
            NoInputReprocessingItemLoader(item=il_loaded).load_item(), {"title": "foo"}
        )

    def test_avoid_reprocessing_with_initial_values_list(self):
        il = NoInputReprocessingItemLoader(
            item=NoInputReprocessingItem(title=["foo", "bar"])
        )
        il_loaded = il.load_item()
        self.assertEqual(il_loaded, {"title": "foo"})
        self.assertEqual(
            NoInputReprocessingItemLoader(item=il_loaded).load_item(), {"title": "foo"}
        )

    def test_avoid_reprocessing_without_initial_values_single(self):
        il = NoInputReprocessingItemLoader()
        il.add_value("title", "FOO")
        il_loaded = il.load_item()
        self.assertEqual(il_loaded, {"title": "FOO"})
        self.assertEqual(
            NoInputReprocessingItemLoader(item=il_loaded).load_item(), {"title": "FOO"}
        )

    def test_avoid_reprocessing_without_initial_values_list(self):
        il = NoInputReprocessingItemLoader()
        il.add_value("title", ["foo", "bar"])
        il_loaded = il.load_item()
        self.assertEqual(il_loaded, {"title": "FOO"})
        self.assertEqual(
            NoInputReprocessingItemLoader(item=il_loaded).load_item(), {"title": "FOO"}
        )


class TestOutputProcessorItem(unittest.TestCase):
    def test_output_processor(self):
        class TempItem(Item):
            temp = Field()

            def __init__(self, *args, **kwargs):
                super().__init__(self, *args, **kwargs)
                self.setdefault("temp", 0.3)

        class TempLoader(ItemLoader):
            default_item_class = TempItem
            default_input_processor = Identity()
            default_output_processor = Compose(TakeFirst())

        loader = TempLoader()
        item = loader.load_item()
        self.assertIsInstance(item, TempItem)
        self.assertEqual(dict(item), {"temp": 0.3})


class SelectortemLoaderTest(unittest.TestCase):
    response = HtmlResponse(
        url="",
        encoding="utf-8",
        body=b"""
    <html>
    <body>
    <div id="id">marta</div>
    <p>paragraph</p>
    <a href="http://www.scrapy.org">homepage</a>
    <img src="/images/logo.png" width="244" height="65" alt="Scrapy">
    </body>
    </html>
    """,
    )

    def test_init_method(self):
        l = TestItemLoader()
        self.assertEqual(l.selector, None)

    def test_init_method_errors(self):
        l = TestItemLoader()
        self.assertRaises(RuntimeError, l.add_xpath, "url", "//a/@href")
        self.assertRaises(RuntimeError, l.replace_xpath, "url", "//a/@href")
        self.assertRaises(RuntimeError, l.get_xpath, "//a/@href")
        self.assertRaises(RuntimeError, l.add_css, "name", "#name::text")
        self.assertRaises(RuntimeError, l.replace_css, "name", "#name::text")
        self.assertRaises(RuntimeError, l.get_css, "#name::text")

    def test_init_method_with_selector(self):
        sel = Selector(text="<html><body><div>marta</div></body></html>")
        l = TestItemLoader(selector=sel)
        self.assertIs(l.selector, sel)

        l.add_xpath("name", "//div/text()")
        self.assertEqual(l.get_output_value("name"), ["Marta"])

    def test_init_method_with_selector_css(self):
        sel = Selector(text="<html><body><div>marta</div></body></html>")
        l = TestItemLoader(selector=sel)
        self.assertIs(l.selector, sel)

        l.add_css("name", "div::text")
        self.assertEqual(l.get_output_value("name"), ["Marta"])

    def test_init_method_with_base_response(self):
        """Selector should be None after initialization"""
        response = Response("https://scrapy.org")
        l = TestItemLoader(response=response)
        self.assertIs(l.selector, None)

    def test_init_method_with_response(self):
        l = TestItemLoader(response=self.response)
        self.assertTrue(l.selector)

        l.add_xpath("name", "//div/text()")
        self.assertEqual(l.get_output_value("name"), ["Marta"])

    def test_init_method_with_response_css(self):
        l = TestItemLoader(response=self.response)
        self.assertTrue(l.selector)

        l.add_css("name", "div::text")
        self.assertEqual(l.get_output_value("name"), ["Marta"])

        l.add_css("url", "a::attr(href)")
        self.assertEqual(l.get_output_value("url"), ["http://www.scrapy.org"])

        # combining/accumulating CSS selectors and XPath expressions
        l.add_xpath("name", "//div/text()")
        self.assertEqual(l.get_output_value("name"), ["Marta", "Marta"])

        l.add_xpath("url", "//img/@src")
        self.assertEqual(
            l.get_output_value("url"), ["http://www.scrapy.org", "/images/logo.png"]
        )

    def test_add_xpath_re(self):
        l = TestItemLoader(response=self.response)
        l.add_xpath("name", "//div/text()", re="ma")
        self.assertEqual(l.get_output_value("name"), ["Ma"])

    def test_replace_xpath(self):
        l = TestItemLoader(response=self.response)
        self.assertTrue(l.selector)
        l.add_xpath("name", "//div/text()")
        self.assertEqual(l.get_output_value("name"), ["Marta"])
        l.replace_xpath("name", "//p/text()")
        self.assertEqual(l.get_output_value("name"), ["Paragraph"])

        l.replace_xpath("name", ["//p/text()", "//div/text()"])
        self.assertEqual(l.get_output_value("name"), ["Paragraph", "Marta"])

    def test_get_xpath(self):
        l = TestItemLoader(response=self.response)
        self.assertEqual(l.get_xpath("//p/text()"), ["paragraph"])
        self.assertEqual(l.get_xpath("//p/text()", TakeFirst()), "paragraph")
        self.assertEqual(l.get_xpath("//p/text()", TakeFirst(), re="pa"), "pa")

        self.assertEqual(
            l.get_xpath(["//p/text()", "//div/text()"]), ["paragraph", "marta"]
        )

    def test_replace_xpath_multi_fields(self):
        l = TestItemLoader(response=self.response)
        l.add_xpath(None, "//div/text()", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(l.get_output_value("name"), ["Marta"])
        l.replace_xpath(None, "//p/text()", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(l.get_output_value("name"), ["Paragraph"])

    def test_replace_xpath_re(self):
        l = TestItemLoader(response=self.response)
        self.assertTrue(l.selector)
        l.add_xpath("name", "//div/text()")
        self.assertEqual(l.get_output_value("name"), ["Marta"])
        l.replace_xpath("name", "//div/text()", re="ma")
        self.assertEqual(l.get_output_value("name"), ["Ma"])

    def test_add_css_re(self):
        l = TestItemLoader(response=self.response)
        l.add_css("name", "div::text", re="ma")
        self.assertEqual(l.get_output_value("name"), ["Ma"])

        l.add_css("url", "a::attr(href)", re="http://(.+)")
        self.assertEqual(l.get_output_value("url"), ["www.scrapy.org"])

    def test_replace_css(self):
        l = TestItemLoader(response=self.response)
        self.assertTrue(l.selector)
        l.add_css("name", "div::text")
        self.assertEqual(l.get_output_value("name"), ["Marta"])
        l.replace_css("name", "p::text")
        self.assertEqual(l.get_output_value("name"), ["Paragraph"])

        l.replace_css("name", ["p::text", "div::text"])
        self.assertEqual(l.get_output_value("name"), ["Paragraph", "Marta"])

        l.add_css("url", "a::attr(href)", re="http://(.+)")
        self.assertEqual(l.get_output_value("url"), ["www.scrapy.org"])
        l.replace_css("url", "img::attr(src)")
        self.assertEqual(l.get_output_value("url"), ["/images/logo.png"])

    def test_get_css(self):
        l = TestItemLoader(response=self.response)
        self.assertEqual(l.get_css("p::text"), ["paragraph"])
        self.assertEqual(l.get_css("p::text", TakeFirst()), "paragraph")
        self.assertEqual(l.get_css("p::text", TakeFirst(), re="pa"), "pa")

        self.assertEqual(l.get_css(["p::text", "div::text"]), ["paragraph", "marta"])
        self.assertEqual(
            l.get_css(["a::attr(href)", "img::attr(src)"]),
            ["http://www.scrapy.org", "/images/logo.png"],
        )

    def test_replace_css_multi_fields(self):
        l = TestItemLoader(response=self.response)
        l.add_css(None, "div::text", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(l.get_output_value("name"), ["Marta"])
        l.replace_css(None, "p::text", TakeFirst(), lambda x: {"name": x})
        self.assertEqual(l.get_output_value("name"), ["Paragraph"])

        l.add_css(None, "a::attr(href)", TakeFirst(), lambda x: {"url": x})
        self.assertEqual(l.get_output_value("url"), ["http://www.scrapy.org"])
        l.replace_css(None, "img::attr(src)", TakeFirst(), lambda x: {"url": x})
        self.assertEqual(l.get_output_value("url"), ["/images/logo.png"])

    def test_replace_css_re(self):
        l = TestItemLoader(response=self.response)
        self.assertTrue(l.selector)
        l.add_css("url", "a::attr(href)")
        self.assertEqual(l.get_output_value("url"), ["http://www.scrapy.org"])
        l.replace_css("url", "a::attr(href)", re=r"http://www\.(.+)")
        self.assertEqual(l.get_output_value("url"), ["scrapy.org"])


class SubselectorLoaderTest(unittest.TestCase):
    response = HtmlResponse(
        url="",
        encoding="utf-8",
        body=b"""
    <html>
    <body>
    <header>
      <div id="id">marta</div>
      <p>paragraph</p>
    </header>
    <footer class="footer">
      <a href="http://www.scrapy.org">homepage</a>
      <img src="/images/logo.png" width="244" height="65" alt="Scrapy">
    </footer>
    </body>
    </html>
    """,
    )

    def test_nested_xpath(self):
        l = NestedItemLoader(response=self.response)

        nl = l.nested_xpath("//header")
        nl.add_xpath("name", "div/text()")
        nl.add_css("name_div", "#id")
        nl.add_value("name_value", nl.selector.xpath('div[@id = "id"]/text()').getall())

        self.assertEqual(l.get_output_value("name"), ["marta"])
        self.assertEqual(l.get_output_value("name_div"), ['<div id="id">marta</div>'])
        self.assertEqual(l.get_output_value("name_value"), ["marta"])

        self.assertEqual(l.get_output_value("name"), nl.get_output_value("name"))
        self.assertEqual(
            l.get_output_value("name_div"), nl.get_output_value("name_div")
        )
        self.assertEqual(
            l.get_output_value("name_value"), nl.get_output_value("name_value")
        )

    def test_nested_css(self):
        l = NestedItemLoader(response=self.response)
        nl = l.nested_css("header")
        nl.add_xpath("name", "div/text()")
        nl.add_css("name_div", "#id")
        nl.add_value("name_value", nl.selector.xpath('div[@id = "id"]/text()').getall())

        self.assertEqual(l.get_output_value("name"), ["marta"])
        self.assertEqual(l.get_output_value("name_div"), ['<div id="id">marta</div>'])
        self.assertEqual(l.get_output_value("name_value"), ["marta"])

        self.assertEqual(l.get_output_value("name"), nl.get_output_value("name"))
        self.assertEqual(
            l.get_output_value("name_div"), nl.get_output_value("name_div")
        )
        self.assertEqual(
            l.get_output_value("name_value"), nl.get_output_value("name_value")
        )

    def test_nested_replace(self):
        l = NestedItemLoader(response=self.response)
        nl1 = l.nested_xpath("//footer")
        nl2 = nl1.nested_xpath("a")

        l.add_xpath("url", "//footer/a/@href")
        self.assertEqual(l.get_output_value("url"), ["http://www.scrapy.org"])
        nl1.replace_xpath("url", "img/@src")
        self.assertEqual(l.get_output_value("url"), ["/images/logo.png"])
        nl2.replace_xpath("url", "@href")
        self.assertEqual(l.get_output_value("url"), ["http://www.scrapy.org"])

    def test_nested_ordering(self):
        l = NestedItemLoader(response=self.response)
        nl1 = l.nested_xpath("//footer")
        nl2 = nl1.nested_xpath("a")

        nl1.add_xpath("url", "img/@src")
        l.add_xpath("url", "//footer/a/@href")
        nl2.add_xpath("url", "text()")
        l.add_xpath("url", "//footer/a/@href")

        self.assertEqual(
            l.get_output_value("url"),
            [
                "/images/logo.png",
                "http://www.scrapy.org",
                "homepage",
                "http://www.scrapy.org",
            ],
        )

    def test_nested_load_item(self):
        l = NestedItemLoader(response=self.response)
        nl1 = l.nested_xpath("//footer")
        nl2 = nl1.nested_xpath("img")

        l.add_xpath("name", "//header/div/text()")
        nl1.add_xpath("url", "a/@href")
        nl2.add_xpath("image", "@src")

        item = l.load_item()

        assert item is l.item
        assert item is nl1.item
        assert item is nl2.item

        self.assertEqual(item["name"], ["marta"])
        self.assertEqual(item["url"], ["http://www.scrapy.org"])
        self.assertEqual(item["image"], ["/images/logo.png"])


# Functions as processors


def function_processor_strip(iterable):
    return [x.strip() for x in iterable]


def function_processor_upper(iterable):
    return [x.upper() for x in iterable]


class FunctionProcessorItem(Item):
    foo = Field(
        input_processor=function_processor_strip,
        output_processor=function_processor_upper,
    )


class FunctionProcessorItemLoader(ItemLoader):
    default_item_class = FunctionProcessorItem


class FunctionProcessorTestCase(unittest.TestCase):
    def test_processor_defined_in_item(self):
        lo = FunctionProcessorItemLoader()
        lo.add_value("foo", "  bar  ")
        lo.add_value("foo", ["  asdf  ", "  qwerty  "])
        self.assertEqual(dict(lo.load_item()), {"foo": ["BAR", "ASDF", "QWERTY"]})


if __name__ == "__main__":
    unittest.main()
